@@ -31,12 +31,13 @@
  *
  * [3] https://github.com/simd-sorting/fast-and-robust: SPDX-License-Identifier: MIT
  *
- * [4] http://mitp-content-server.mit.edu:18180/books/content/sectbyfn?collid=books_pres_0&fn=Chapter%2027.pdf&id=8030
+ * [4] https://mitp-content-server.mit.edu/books/content/sectbyfn?collid=books_pres_0&fn=Chapter%2027.pdf&id=8030
  *
  */
-#include "xss-common-includes.h"
 #include "xss-pivot-selection.hpp"
 #include "xss-network-qsort.hpp"
+#include "xss-common-comparators.hpp"
+#include "xss-reg-networks.hpp"
 
 template <typename T>
 bool is_a_nan(T elem)
@@ -87,7 +88,8 @@ X86_SIMD_SORT_INLINE bool array_has_nan(type_t *arr, arrsize_t size)
         else {
             in = vtype::loadu(arr + ii);
         }
-        auto nanmask = vtype::convert_mask_to_int(vtype::template fpclass<0x01 | 0x80>(in));
+        auto nanmask = vtype::convert_mask_to_int(
+                vtype::template fpclass<0x01 | 0x80>(in));
         if (nanmask != 0x00) {
             found_nan = true;
             break;
@@ -97,17 +99,32 @@ X86_SIMD_SORT_INLINE bool array_has_nan(type_t *arr, arrsize_t size)
 }
 
 template <typename type_t>
-X86_SIMD_SORT_INLINE void
-replace_inf_with_nan(type_t *arr, arrsize_t size, arrsize_t nan_count)
+X86_SIMD_SORT_INLINE void replace_inf_with_nan(type_t *arr,
+                                               arrsize_t size,
+                                               arrsize_t nan_count,
+                                               bool descending = false)
 {
-    for (arrsize_t ii = size - 1; nan_count > 0; --ii) {
-        if constexpr (std::is_floating_point_v<type_t>) {
-            arr[ii] = std::numeric_limits<type_t>::quiet_NaN();
+    if (descending) {
+        for (arrsize_t ii = 0; nan_count > 0; ++ii) {
+            if constexpr (xss::fp::is_floating_point_v<type_t>) {
+                arr[ii] = xss::fp::quiet_NaN<type_t>();
+            }
+            else {
+                arr[ii] = 0xFFFF;
+            }
+            nan_count -= 1;
         }
-        else {
-            arr[ii] = 0xFFFF;
+    }
+    else {
+        for (arrsize_t ii = size - 1; nan_count > 0; --ii) {
+            if constexpr (xss::fp::is_floating_point_v<type_t>) {
+                arr[ii] = xss::fp::quiet_NaN<type_t>();
+            }
+            else {
+                arr[ii] = 0xFFFF;
+            }
+            nan_count -= 1;
         }
-        nan_count -= 1;
     }
 }
 
@@ -136,7 +153,27 @@ X86_SIMD_SORT_INLINE arrsize_t move_nans_to_end_of_array(T *arr, arrsize_t size)
     return size - count - 1;
 }
 
-template <typename vtype, typename T = typename vtype::type_t>
+/*
+ * Sort all the NAN's to start of the array and return the index of the first elem
+ * in the array which is not a nan
+ */
+template <typename T>
+X86_SIMD_SORT_INLINE arrsize_t move_nans_to_start_of_array(T *arr,
+                                                           arrsize_t size)
+{
+    arrsize_t count = 0;
+
+    for (arrsize_t i = 0; i < size; i++) {
+        if (is_a_nan(arr[i])) {
+            std::swap(arr[count], arr[i]);
+            count++;
+        }
+    }
+
+    return count;
+}
+
+template <typename vtype, typename T>
 X86_SIMD_SORT_INLINE bool comparison_func(const T &a, const T &b)
 {
     return a < b;
@@ -180,6 +217,7 @@ int avx512_double_compressstore(type_t *left_addr,
 
 // Generic function dispatches to AVX2 or AVX512 code
 template <typename vtype,
+          typename comparator,
           typename type_t,
           typename reg_t = typename vtype::reg_t>
 X86_SIMD_SORT_INLINE arrsize_t partition_vec(type_t *l_store,
@@ -189,10 +227,11 @@ X86_SIMD_SORT_INLINE arrsize_t partition_vec(type_t *l_store,
                                              reg_t &smallest_vec,
                                              reg_t &biggest_vec)
 {
-    typename vtype::opmask_t ge_mask = vtype::ge(curr_vec, pivot_vec);
+    typename vtype::opmask_t right_mask
+            = comparator::PartitionComparator(curr_vec, pivot_vec);
 
-    int amount_ge_pivot
-            = vtype::double_compressstore(l_store, r_store, ge_mask, curr_vec);
+    int amount_ge_pivot = vtype::double_compressstore(
+            l_store, r_store, right_mask, curr_vec);
 
     smallest_vec = vtype::min(curr_vec, smallest_vec);
     biggest_vec = vtype::max(curr_vec, biggest_vec);
@@ -204,19 +243,19 @@ X86_SIMD_SORT_INLINE arrsize_t partition_vec(type_t *l_store,
  * Parition an array based on the pivot and returns the index of the
  * first element that is greater than or equal to the pivot.
  */
-template <typename vtype, typename type_t>
-X86_SIMD_SORT_INLINE arrsize_t partition_avx512(type_t *arr,
-                                                arrsize_t left,
-                                                arrsize_t right,
-                                                type_t pivot,
-                                                type_t *smallest,
-                                                type_t *biggest)
+template <typename vtype, typename comparator, typename type_t>
+X86_SIMD_SORT_INLINE arrsize_t partition(type_t *arr,
+                                         arrsize_t left,
+                                         arrsize_t right,
+                                         type_t pivot,
+                                         type_t *smallest,
+                                         type_t *biggest)
 {
     /* make array length divisible by vtype::numlanes , shortening the array */
     for (int32_t i = (right - left) % vtype::numlanes; i > 0; --i) {
         *smallest = std::min(*smallest, arr[left], comparison_func<vtype>);
         *biggest = std::max(*biggest, arr[left], comparison_func<vtype>);
-        if (!comparison_func<vtype>(arr[left], pivot)) {
+        if (!comparator::STDSortComparator(arr[left], pivot)) {
             std::swap(arr[left], arr[--right]);
         }
         else {
@@ -237,13 +276,13 @@ X86_SIMD_SORT_INLINE arrsize_t partition_avx512(type_t *arr,
         arrsize_t unpartitioned = right - left - vtype::numlanes;
         arrsize_t l_store = left;
 
-        arrsize_t amount_ge_pivot
-                = partition_vec<vtype>(arr + l_store,
-                                       arr + l_store + unpartitioned,
-                                       vec,
-                                       pivot_vec,
-                                       min_vec,
-                                       max_vec);
+        arrsize_t amount_ge_pivot = partition_vec<vtype, comparator>(
+                arr + l_store,
+                arr + l_store + unpartitioned,
+                vec,
+                pivot_vec,
+                min_vec,
+                max_vec);
         l_store += (vtype::numlanes - amount_ge_pivot);
         *smallest = vtype::reducemin(min_vec);
         *biggest = vtype::reducemax(max_vec);
@@ -276,34 +315,35 @@ X86_SIMD_SORT_INLINE arrsize_t partition_avx512(type_t *arr,
             left += vtype::numlanes;
         }
         // partition the current vector and save it on both sides of the array
-        arrsize_t amount_ge_pivot
-                = partition_vec<vtype>(arr + l_store,
-                                       arr + l_store + unpartitioned,
-                                       curr_vec,
-                                       pivot_vec,
-                                       min_vec,
-                                       max_vec);
+        arrsize_t amount_ge_pivot = partition_vec<vtype, comparator>(
+                arr + l_store,
+                arr + l_store + unpartitioned,
+                curr_vec,
+                pivot_vec,
+                min_vec,
+                max_vec);
         l_store += (vtype::numlanes - amount_ge_pivot);
         unpartitioned -= vtype::numlanes;
     }
 
     /* partition and save vec_left and vec_right */
     arrsize_t amount_ge_pivot
-            = partition_vec<vtype>(arr + l_store,
-                                   arr + l_store + unpartitioned,
-                                   vec_left,
-                                   pivot_vec,
-                                   min_vec,
-                                   max_vec);
+            = partition_vec<vtype, comparator>(arr + l_store,
+                                               arr + l_store + unpartitioned,
+                                               vec_left,
+                                               pivot_vec,
+                                               min_vec,
+                                               max_vec);
     l_store += (vtype::numlanes - amount_ge_pivot);
     unpartitioned -= vtype::numlanes;
 
-    amount_ge_pivot = partition_vec<vtype>(arr + l_store,
-                                           arr + l_store + unpartitioned,
-                                           vec_right,
-                                           pivot_vec,
-                                           min_vec,
-                                           max_vec);
+    amount_ge_pivot
+            = partition_vec<vtype, comparator>(arr + l_store,
+                                               arr + l_store + unpartitioned,
+                                               vec_right,
+                                               pivot_vec,
+                                               min_vec,
+                                               max_vec);
     l_store += (vtype::numlanes - amount_ge_pivot);
     unpartitioned -= vtype::numlanes;
 
@@ -313,23 +353,24 @@ X86_SIMD_SORT_INLINE arrsize_t partition_avx512(type_t *arr,
 }
 
 template <typename vtype,
+          typename comparator,
           int num_unroll,
           typename type_t = typename vtype::type_t>
-X86_SIMD_SORT_INLINE arrsize_t partition_avx512_unrolled(type_t *arr,
-                                                         arrsize_t left,
-                                                         arrsize_t right,
-                                                         type_t pivot,
-                                                         type_t *smallest,
-                                                         type_t *biggest)
+X86_SIMD_SORT_INLINE arrsize_t partition_unrolled(type_t *arr,
+                                                  arrsize_t left,
+                                                  arrsize_t right,
+                                                  type_t pivot,
+                                                  type_t *smallest,
+                                                  type_t *biggest)
 {
     if constexpr (num_unroll == 0) {
-        return partition_avx512<vtype>(
+        return partition<vtype, comparator>(
                 arr, left, right, pivot, smallest, biggest);
     }
 
-    /* Use regular partition_avx512 for smaller arrays */
+    /* Use regular partition for smaller arrays */
     if (right - left < 3 * num_unroll * vtype::numlanes) {
-        return partition_avx512<vtype>(
+        return partition<vtype, comparator>(
                 arr, left, right, pivot, smallest, biggest);
     }
 
@@ -337,7 +378,7 @@ X86_SIMD_SORT_INLINE arrsize_t partition_avx512_unrolled(type_t *arr,
     for (int32_t i = ((right - left) % (vtype::numlanes)); i > 0; --i) {
         *smallest = std::min(*smallest, arr[left], comparison_func<vtype>);
         *biggest = std::max(*biggest, arr[left], comparison_func<vtype>);
-        if (!comparison_func<vtype>(arr[left], pivot)) {
+        if (!comparator::STDSortComparator(arr[left], pivot)) {
             std::swap(arr[left], arr[--right]);
         }
         else {
@@ -419,13 +460,13 @@ X86_SIMD_SORT_INLINE arrsize_t partition_avx512_unrolled(type_t *arr,
          * */
         X86_SIMD_SORT_UNROLL_LOOP(8)
         for (int ii = 0; ii < num_unroll; ++ii) {
-            arrsize_t amount_ge_pivot
-                    = partition_vec<vtype>(arr + l_store,
-                                           arr + l_store + unpartitioned,
-                                           curr_vec[ii],
-                                           pivot_vec,
-                                           min_vec,
-                                           max_vec);
+            arrsize_t amount_ge_pivot = partition_vec<vtype, comparator>(
+                    arr + l_store,
+                    arr + l_store + unpartitioned,
+                    curr_vec[ii],
+                    pivot_vec,
+                    min_vec,
+                    max_vec);
             l_store += (vtype::numlanes - amount_ge_pivot);
             unpartitioned -= vtype::numlanes;
         }
@@ -434,25 +475,25 @@ X86_SIMD_SORT_INLINE arrsize_t partition_avx512_unrolled(type_t *arr,
     /* partition and save vec_left[num_unroll] and vec_right[num_unroll] */
     X86_SIMD_SORT_UNROLL_LOOP(8)
     for (int ii = 0; ii < num_unroll; ++ii) {
-        arrsize_t amount_ge_pivot
-                = partition_vec<vtype>(arr + l_store,
-                                       arr + l_store + unpartitioned,
-                                       vec_left[ii],
-                                       pivot_vec,
-                                       min_vec,
-                                       max_vec);
+        arrsize_t amount_ge_pivot = partition_vec<vtype, comparator>(
+                arr + l_store,
+                arr + l_store + unpartitioned,
+                vec_left[ii],
+                pivot_vec,
+                min_vec,
+                max_vec);
         l_store += (vtype::numlanes - amount_ge_pivot);
         unpartitioned -= vtype::numlanes;
     }
     X86_SIMD_SORT_UNROLL_LOOP(8)
     for (int ii = 0; ii < num_unroll; ++ii) {
-        arrsize_t amount_ge_pivot
-                = partition_vec<vtype>(arr + l_store,
-                                       arr + l_store + unpartitioned,
-                                       vec_right[ii],
-                                       pivot_vec,
-                                       min_vec,
-                                       max_vec);
+        arrsize_t amount_ge_pivot = partition_vec<vtype, comparator>(
+                arr + l_store,
+                arr + l_store + unpartitioned,
+                vec_right[ii],
+                pivot_vec,
+                min_vec,
+                max_vec);
         l_store += (vtype::numlanes - amount_ge_pivot);
         unpartitioned -= vtype::numlanes;
     }
@@ -460,13 +501,13 @@ X86_SIMD_SORT_INLINE arrsize_t partition_avx512_unrolled(type_t *arr,
     /* partition and save vec_align[vecsToPartition] */
     X86_SIMD_SORT_UNROLL_LOOP(8)
     for (int ii = 0; ii < vecsToPartition; ++ii) {
-        arrsize_t amount_ge_pivot
-                = partition_vec<vtype>(arr + l_store,
-                                       arr + l_store + unpartitioned,
-                                       vec_align[ii],
-                                       pivot_vec,
-                                       min_vec,
-                                       max_vec);
+        arrsize_t amount_ge_pivot = partition_vec<vtype, comparator>(
+                arr + l_store,
+                arr + l_store + unpartitioned,
+                vec_align[ii],
+                pivot_vec,
+                min_vec,
+                max_vec);
         l_store += (vtype::numlanes - amount_ge_pivot);
         unpartitioned -= vtype::numlanes;
     }
@@ -479,40 +520,85 @@ X86_SIMD_SORT_INLINE arrsize_t partition_avx512_unrolled(type_t *arr,
 template <typename vtype, int maxN>
 void sort_n(typename vtype::type_t *arr, int N);
 
-template <typename vtype, typename type_t>
-static void
-qsort_(type_t *arr, arrsize_t left, arrsize_t right, arrsize_t max_iters)
+template <typename vtype, typename comparator, typename type_t>
+static void qsort_(type_t *arr,
+                   arrsize_t left,
+                   arrsize_t right,
+                   arrsize_t max_iters,
+                   arrsize_t task_threshold)
 {
     /*
      * Resort to std::sort if quicksort isnt making any progress
      */
     if (max_iters <= 0) {
-        std::sort(arr + left, arr + right + 1, comparison_func<vtype>);
+        std::sort(arr + left, arr + right + 1, comparator::STDSortComparator);
         return;
     }
     /*
      * Base case: use bitonic networks to sort arrays <= vtype::network_sort_threshold
      */
     if (right + 1 - left <= vtype::network_sort_threshold) {
-        sort_n<vtype, vtype::network_sort_threshold>(
+        sort_n<vtype, comparator, vtype::network_sort_threshold>(
                 arr + left, (int32_t)(right + 1 - left));
         return;
     }
 
-    type_t pivot = get_pivot_blocks<vtype, type_t>(arr, left, right);
+    auto pivot_result
+            = get_pivot_smart<vtype, comparator, type_t>(arr, left, right);
+    type_t pivot = pivot_result.pivot;
+
+    if (pivot_result.result == pivot_result_t::Sorted) { return; }
+
     type_t smallest = vtype::type_max();
     type_t biggest = vtype::type_min();
 
-    arrsize_t pivot_index
-            = partition_avx512_unrolled<vtype, vtype::partition_unroll_factor>(
-                    arr, left, right + 1, pivot, &smallest, &biggest);
+    arrsize_t pivot_index = partition_unrolled<vtype,
+                                               comparator,
+                                               vtype::partition_unroll_factor>(
+            arr, left, right + 1, pivot, &smallest, &biggest);
 
-    if (pivot != smallest)
-        qsort_<vtype>(arr, left, pivot_index - 1, max_iters - 1);
-    if (pivot != biggest) qsort_<vtype>(arr, pivot_index, right, max_iters - 1);
+    if (pivot_result.result == pivot_result_t::Only2Values) { return; }
+
+    type_t leftmostValue = comparator::leftmost(smallest, biggest);
+    type_t rightmostValue = comparator::rightmost(smallest, biggest);
+
+#ifdef XSS_COMPILE_OPENMP
+    if (pivot != leftmostValue) {
+        bool parallel_left = (pivot_index - left) > task_threshold;
+        if (parallel_left) {
+#pragma omp task
+            qsort_<vtype, comparator>(
+                    arr, left, pivot_index - 1, max_iters - 1, task_threshold);
+        }
+        else {
+            qsort_<vtype, comparator>(
+                    arr, left, pivot_index - 1, max_iters - 1, task_threshold);
+        }
+    }
+    if (pivot != rightmostValue) {
+        bool parallel_right = (right - pivot_index) > task_threshold;
+
+        if (parallel_right) {
+#pragma omp task
+            qsort_<vtype, comparator>(
+                    arr, pivot_index, right, max_iters - 1, task_threshold);
+        }
+        else {
+            qsort_<vtype, comparator>(
+                    arr, pivot_index, right, max_iters - 1, task_threshold);
+        }
+    }
+#else
+    UNUSED(task_threshold);
+
+    if (pivot != leftmostValue)
+        qsort_<vtype, comparator>(arr, left, pivot_index - 1, max_iters - 1, 0);
+    if (pivot != rightmostValue)
+        qsort_<vtype, comparator>(arr, pivot_index, right, max_iters - 1, 0);
+#endif
 }
 
-template <typename vtype, typename type_t>
+template <typename vtype, typename comparator, typename type_t>
 X86_SIMD_SORT_INLINE void qselect_(type_t *arr,
                                    arrsize_t pos,
                                    arrsize_t left,
@@ -523,97 +609,191 @@ X86_SIMD_SORT_INLINE void qselect_(type_t *arr,
      * Resort to std::sort if quicksort isnt making any progress
      */
     if (max_iters <= 0) {
-        std::sort(arr + left, arr + right + 1, comparison_func<vtype>);
+        std::sort(arr + left, arr + right + 1, comparator::STDSortComparator);
         return;
     }
     /*
      * Base case: use bitonic networks to sort arrays <= vtype::network_sort_threshold
      */
     if (right + 1 - left <= vtype::network_sort_threshold) {
-        sort_n<vtype, vtype::network_sort_threshold>(
+        sort_n<vtype, comparator, vtype::network_sort_threshold>(
                 arr + left, (int32_t)(right + 1 - left));
         return;
     }
 
     type_t pivot = get_pivot<vtype, type_t>(arr, left, right);
+
     type_t smallest = vtype::type_max();
     type_t biggest = vtype::type_min();
 
-    arrsize_t pivot_index
-            = partition_avx512_unrolled<vtype, vtype::partition_unroll_factor>(
-                    arr, left, right + 1, pivot, &smallest, &biggest);
+    arrsize_t pivot_index = partition_unrolled<vtype,
+                                               comparator,
+                                               vtype::partition_unroll_factor>(
+            arr, left, right + 1, pivot, &smallest, &biggest);
 
-    if ((pivot != smallest) && (pos < pivot_index))
-        qselect_<vtype>(arr, pos, left, pivot_index - 1, max_iters - 1);
-    else if ((pivot != biggest) && (pos >= pivot_index))
-        qselect_<vtype>(arr, pos, pivot_index, right, max_iters - 1);
+    type_t leftmostValue = comparator::leftmost(smallest, biggest);
+    type_t rightmostValue = comparator::rightmost(smallest, biggest);
+
+    if ((pivot != leftmostValue) && (pos < pivot_index))
+        qselect_<vtype, comparator>(
+                arr, pos, left, pivot_index - 1, max_iters - 1);
+    else if ((pivot != rightmostValue) && (pos >= pivot_index))
+        qselect_<vtype, comparator>(
+                arr, pos, pivot_index, right, max_iters - 1);
 }
 
 // Quicksort routines:
-template <typename vtype, typename T>
+template <typename vtype, typename T, bool descending = false>
 X86_SIMD_SORT_INLINE void xss_qsort(T *arr, arrsize_t arrsize, bool hasnan)
 {
+    using comparator =
+            typename std::conditional<descending,
+                                      Comparator<vtype, true>,
+                                      Comparator<vtype, false>>::type;
+
     if (arrsize > 1) {
-        if constexpr (std::is_floating_point_v<T>) {
-            arrsize_t nan_count = 0;
+        arrsize_t nan_count = 0;
+        if constexpr (xss::fp::is_floating_point_v<T>) {
             if (UNLIKELY(hasnan)) {
                 nan_count = replace_nan_with_inf<vtype>(arr, arrsize);
             }
-            qsort_<vtype, T>(arr, 0, arrsize - 1, 2 * (arrsize_t)log2(arrsize));
-            replace_inf_with_nan(arr, arrsize, nan_count);
+        }
+
+        UNUSED(hasnan);
+
+#ifdef XSS_COMPILE_OPENMP
+
+        bool use_parallel = arrsize > 100000;
+
+        if (use_parallel) {
+            // This thread limit was determined experimentally; it may be better for it to be the number of physical cores on the system
+            constexpr int thread_limit = 8;
+            int thread_count = std::min(thread_limit, omp_get_max_threads());
+            arrsize_t task_threshold
+                    = std::max((arrsize_t)100000, arrsize / 100);
+
+            // We use omp parallel and then omp single to setup the threads that will run the omp task calls in qsort_
+            // The omp single prevents multiple threads from running the initial qsort_ simultaneously and causing problems
+            // Note that we do not use the if(...) clause built into OpenMP, because it causes a performance regression for small arrays
+#pragma omp parallel num_threads(thread_count)
+#pragma omp single
+            qsort_<vtype, comparator, T>(arr,
+                                         0,
+                                         arrsize - 1,
+                                         2 * (arrsize_t)log2(arrsize),
+                                         task_threshold);
+#pragma omp taskwait
         }
         else {
-            UNUSED(hasnan);
-            qsort_<vtype, T>(arr, 0, arrsize - 1, 2 * (arrsize_t)log2(arrsize));
+            qsort_<vtype, comparator, T>(arr,
+                                         0,
+                                         arrsize - 1,
+                                         2 * (arrsize_t)log2(arrsize),
+                                         std::numeric_limits<arrsize_t>::max());
         }
+#else
+        qsort_<vtype, comparator, T>(
+                arr, 0, arrsize - 1, 2 * (arrsize_t)log2(arrsize), 0);
+#endif
+
+        replace_inf_with_nan(arr, arrsize, nan_count, descending);
     }
+
+#ifdef __MMX__
+    // Workaround for compiler bug generating MMX instructions without emms
+    _mm_empty();
+#endif
 }
 
 // Quick select methods
-template <typename vtype, typename T>
+template <typename vtype, typename T, bool descending = false>
 X86_SIMD_SORT_INLINE void
 xss_qselect(T *arr, arrsize_t k, arrsize_t arrsize, bool hasnan)
 {
-    arrsize_t indx_last_elem = arrsize - 1;
-    if constexpr (std::is_floating_point_v<T>) {
+    using comparator =
+            typename std::conditional<descending,
+                                      Comparator<vtype, true>,
+                                      Comparator<vtype, false>>::type;
+
+    // Exit early if no work would be done
+    if (arrsize <= 1) return;
+
+    arrsize_t index_first_elem = 0;
+    arrsize_t index_last_elem = arrsize - 1;
+
+    if constexpr (xss::fp::is_floating_point_v<T>) {
         if (UNLIKELY(hasnan)) {
-            indx_last_elem = move_nans_to_end_of_array(arr, arrsize);
+            if constexpr (descending) {
+                index_first_elem = move_nans_to_start_of_array(arr, arrsize);
+            }
+            else {
+                index_last_elem = move_nans_to_end_of_array(arr, arrsize);
+            }
         }
     }
+
     UNUSED(hasnan);
-    if (indx_last_elem >= k) {
-        qselect_<vtype, T>(
-                arr, k, 0, indx_last_elem, 2 * (arrsize_t)log2(indx_last_elem));
+    if (index_first_elem <= k && index_last_elem >= k) {
+        qselect_<vtype, comparator, T>(arr,
+                                       k,
+                                       index_first_elem,
+                                       index_last_elem,
+                                       2 * (arrsize_t)log2(arrsize));
     }
+
+#ifdef __MMX__
+    // Workaround for compiler bug generating MMX instructions without emms
+    _mm_empty();
+#endif
 }
 
 // Partial sort methods:
-template <typename vtype, typename T>
+template <typename vtype, typename T, bool descending = false>
 X86_SIMD_SORT_INLINE void
 xss_partial_qsort(T *arr, arrsize_t k, arrsize_t arrsize, bool hasnan)
 {
-    xss_qselect<vtype, T>(arr, k - 1, arrsize, hasnan);
-    xss_qsort<vtype, T>(arr, k - 1, hasnan);
+    if (k == 0) return;
+    xss_qselect<vtype, T, descending>(arr, k - 1, arrsize, hasnan);
+    xss_qsort<vtype, T, descending>(arr, k - 1, hasnan);
 }
 
 #define DEFINE_METHODS(ISA, VTYPE) \
     template <typename T> \
-    X86_SIMD_SORT_INLINE void ISA##_qsort( \
-            T *arr, arrsize_t size, bool hasnan = false) \
+    X86_SIMD_SORT_INLINE void ISA##_qsort(T *arr, \
+                                          arrsize_t size, \
+                                          bool hasnan = false, \
+                                          bool descending = false) \
     { \
-        xss_qsort<VTYPE, T>(arr, size, hasnan); \
+        if (descending) { xss_qsort<VTYPE, T, true>(arr, size, hasnan); } \
+        else { \
+            xss_qsort<VTYPE, T, false>(arr, size, hasnan); \
+        } \
     } \
     template <typename T> \
-    X86_SIMD_SORT_INLINE void ISA##_qselect( \
-            T *arr, arrsize_t k, arrsize_t size, bool hasnan = false) \
+    X86_SIMD_SORT_INLINE void ISA##_qselect(T *arr, \
+                                            arrsize_t k, \
+                                            arrsize_t size, \
+                                            bool hasnan = false, \
+                                            bool descending = false) \
     { \
-        xss_qselect<VTYPE, T>(arr, k, size, hasnan); \
+        if (descending) { xss_qselect<VTYPE, T, true>(arr, k, size, hasnan); } \
+        else { \
+            xss_qselect<VTYPE, T, false>(arr, k, size, hasnan); \
+        } \
     } \
     template <typename T> \
-    X86_SIMD_SORT_INLINE void ISA##_partial_qsort( \
-            T *arr, arrsize_t k, arrsize_t size, bool hasnan = false) \
+    X86_SIMD_SORT_INLINE void ISA##_partial_qsort(T *arr, \
+                                                  arrsize_t k, \
+                                                  arrsize_t size, \
+                                                  bool hasnan = false, \
+                                                  bool descending = false) \
     { \
-        xss_partial_qsort<VTYPE, T>(arr, k, size, hasnan); \
+        if (descending) { \
+            xss_partial_qsort<VTYPE, T, true>(arr, k, size, hasnan); \
+        } \
+        else { \
+            xss_partial_qsort<VTYPE, T, false>(arr, k, size, hasnan); \
+        } \
     }
 
 DEFINE_METHODS(avx512, zmm_vector<T>)
