@@ -1,15 +1,9 @@
 """A module containing the `_NestedSequence` protocol."""
 
-from __future__ import annotations
+from typing import TYPE_CHECKING, Any, Protocol, TypeVar, runtime_checkable
 
-from collections.abc import Iterator
-from typing import (
-    Any,
-    overload,
-    TypeVar,
-    Protocol,
-    runtime_checkable,
-)
+if TYPE_CHECKING:
+    from collections.abc import Iterator
 
 __all__ = ["_NestedSequence"]
 
@@ -34,8 +28,6 @@ class _NestedSequence(Protocol[_T_co]):
     --------
     .. code-block:: python
 
-        >>> from __future__ import annotations
-
         >>> from typing import TYPE_CHECKING
         >>> import numpy as np
         >>> from numpy._typing import _NestedSequence
@@ -62,12 +54,7 @@ class _NestedSequence(Protocol[_T_co]):
         """Implement ``len(self)``."""
         raise NotImplementedError
 
-    @overload
-    def __getitem__(self, index: int, /) -> _T_co | _NestedSequence[_T_co]: ...
-    @overload
-    def __getitem__(self, index: slice, /) -> _NestedSequence[_T_co]: ...
-
-    def __getitem__(self, index, /):
+    def __getitem__(self, index: int, /) -> "_T_co | _NestedSequence[_T_co]":
         """Implement ``self[x]``."""
         raise NotImplementedError
 
@@ -75,11 +62,11 @@ class _NestedSequence(Protocol[_T_co]):
         """Implement ``x in self``."""
         raise NotImplementedError
 
-    def __iter__(self, /) -> Iterator[_T_co | _NestedSequence[_T_co]]:
+    def __iter__(self, /) -> "Iterator[_T_co | _NestedSequence[_T_co]]":
         """Implement ``iter(self)``."""
         raise NotImplementedError
 
-    def __reversed__(self, /) -> Iterator[_T_co | _NestedSequence[_T_co]]:
+    def __reversed__(self, /) -> "Iterator[_T_co | _NestedSequence[_T_co]]":
         """Implement ``reversed(self)``."""
         raise NotImplementedError
 
