@@ -1023,7 +1023,7 @@ RAND_INT_TYPE random_zipf(bitgen_t *bitgen_state, double a) {
    * Values below Umin would result in X being rejected because it is too
    * large, so there is no point in including them in the distribution of U.
    */
-  Umin = pow(RAND_INT_MAX, -am1);
+  Umin = pow((double) RAND_INT_MAX, -am1);
   while (1) {
     double U01, T, U, V, X;
 
